"""Module will not import due to invalid blahblah import."""
import blahblah.blah


print("I dont do much")


def arbmodinvalid_attribute():
    """Can't ever run on account of how blah doesn't exist."""
    blahblah.blah.blah()
